"""This module provides a CLI util to make updates to normalizer database."""
import click
from botocore.exceptions import ClientError
from therapy import ACCEPTED_SOURCES
from therapy.etl.merge import Merge
from timeit import default_timer as timer
from boto3.dynamodb.conditions import Key
from therapy.schemas import SourceName
from therapy import SOURCES_CLASS, SOURCES
from therapy.database import Database
from disease.database import Database as DiseaseDatabase
from disease.cli import CLI as DiseaseCLI
from disease.schemas import SourceName as DiseaseSources
from os import environ
import logging

logger = logging.getLogger('therapy')
logger.setLevel(logging.DEBUG)


class CLI:
    """Class for updating the normalizer database via Click"""

    @staticmethod
    @click.command()
    @click.option(
        '--normalizer',
        help="The normalizer(s) you wish to update separated by spaces."
    )
    @click.option(
        '--prod',
        is_flag=True,
        help="Working in production environment."
    )
    @click.option(
        '--db_url',
        help="URL endpoint for the application database."
    )
    @click.option(
        '--update_all',
        is_flag=True,
        help='Update all normalizer sources.'
    )
    @click.option(
        '--update_merged',
        is_flag=True,
        help='Update concepts for normalize endpoint from accepted sources.'
    )
    def update_normalizer_db(normalizer, prod, db_url, update_all,
                             update_merged):
        """Update select normalizer source(s) in the therapy database."""
        endpoint_url = None
        if prod:
            environ['THERAPY_NORM_PROD'] = "TRUE"
            environ['DISEASE_NORM_PROD'] = "TRUE"
            db: Database = Database()
        else:
            if db_url:
                endpoint_url = db_url
            elif 'THERAPY_NORM_DB_URL' in environ.keys():
                endpoint_url = environ['THERAPY_NORM_DB_URL']
            else:
                endpoint_url = 'http://localhost:8000'
            db: Database = Database(db_url=endpoint_url)

        if update_all:
            normalizers = list(src for src in SOURCES)
            CLI()._check_disease_normalizer(normalizers, endpoint_url)
            CLI()._update_normalizers(normalizers, db, update_merged)
        elif not normalizer:
            CLI()._help_msg()
        else:
            normalizers = str(normalizer).lower().split()

            if len(normalizers) == 0:
                CLI()._help_msg()

            non_sources = set(normalizers) - {src for src in SOURCES}

            if len(non_sources) != 0:
                raise Exception(f"Not valid source(s): {non_sources}")

            CLI()._check_disease_normalizer(normalizers, endpoint_url)
            CLI()._update_normalizers(normalizers, db, update_merged)

    def _check_disease_normalizer(self, normalizers, endpoint_url):
        """Load Disease Normalizer data if Hemonc source included.

        :param list normalizers: List of sources to load
        :param str endpoint_url: Therapy endpoint URL
        """
        if 'hemonc' in normalizers:
            db = DiseaseDatabase(db_url=endpoint_url)
            n_sources = len({v.value for v in DiseaseSources})
            if db.diseases.item_count == 0 or \
                    db.metadata.item_count != n_sources:
                msg = "Disease Normalizer not loaded. " \
                      "Loading Disease Normalizer..."
                logger.debug(msg)
                click.echo(msg)
                # Is there a better way to do this?
                try:
                    DiseaseCLI().update_normalizer_db(
                        ['--update_all', '--update_merged',
                         '--db_url', endpoint_url]
                    )
                except Exception as e:
                    logger.error(e)
                    raise Exception(e)
                except:  # noqa: E722
                    pass

    @staticmethod
    def _help_msg():
        """Display help message."""
        ctx = click.get_current_context()
        click.echo(
            "Must either enter 1 or more sources, or use `--update_all` parameter")  # noqa: E501
        click.echo(ctx.get_help())
        ctx.exit()

    @staticmethod
    def _update_normalizers(normalizers, db, update_merged):
        """Update selected normalizer sources."""
        processed_ids = list()
        for n in normalizers:
            msg = f"Deleting {n}..."
            click.echo(f"\n{msg}")
            logger.info(msg)
            start_delete = timer()
            CLI()._delete_data(n, db)
            end_delete = timer()
            delete_time = end_delete - start_delete
            msg = f"Deleted {n} in {delete_time:.5f} seconds."
            click.echo(f"{msg}\n")
            logger.info(msg)

            msg = f"Loading {n}..."
            click.echo(msg)
            logger.info(msg)
            start_load = timer()
            source = SOURCES_CLASS[n](database=db)
            if source.in_normalize:
                processed_ids += source.perform_etl()
            else:
                source.perform_etl()
            end_load = timer()
            load_time = end_load - start_load
            msg = f"Loaded {n} in {load_time:.5f} seconds."
            click.echo(msg)
            logger.info(msg)
            msg = f"Total time for {n}: " \
                  f"{(delete_time + load_time):.5f} seconds."
            click.echo(msg)
            logger.info(msg)

        if update_merged:
            start_merge = timer()
            if not processed_ids or not ACCEPTED_SOURCES.issubset(normalizers):
                CLI()._delete_normalized_data(db)
                processed_ids = db.get_ids_for_merge()
            merge = Merge(database=db)
            click.echo("Constructing normalized records...")
            merge.create_merged_concepts(processed_ids)
            end_merge = timer()
            click.echo(f"Merged concept generation completed in"
                       f" {(end_merge - start_merge):.5f} seconds.")

    @staticmethod
    def _delete_normalized_data(database):
        click.echo("\nDeleting normalized records...")
        start_delete = timer()
        try:
            while True:
                with database.therapies.batch_writer(
                        overwrite_by_pkeys=['label_and_type', 'concept_id']) \
                        as batch:
                    response = database.therapies.query(
                        IndexName='item_type_index',
                        KeyConditionExpression=Key('item_type').eq('merger'),
                    )
                    records = response['Items']
                    if not records:
                        break
                    for record in records:
                        batch.delete_item(Key={
                            'label_and_type': record['label_and_type'],
                            'concept_id': record['concept_id']
                        })
        except ClientError as e:
            click.echo(e.response['Error']['Message'])
        end_delete = timer()
        delete_time = end_delete - start_delete
        click.echo(f"Deleted normalized records in {delete_time:.5f} seconds.")

    @staticmethod
    def _delete_data(source, database):
        # Delete source's metadata
        try:
            metadata = database.metadata.query(
                KeyConditionExpression=Key(
                    'src_name').eq(SourceName[f"{source.upper()}"].value)
            )
            if metadata['Items']:
                database.metadata.delete_item(
                    Key={'src_name': metadata['Items'][0]['src_name']},
                    ConditionExpression="src_name = :src",
                    ExpressionAttributeValues={
                        ':src': SourceName[f"{source.upper()}"].value}
                )
        except ClientError as e:
            click.echo(e.response['Error']['Message'])

        try:
            while True:
                response = database.therapies.query(
                    IndexName='src_index',
                    KeyConditionExpression=Key('src_name').eq(
                        SourceName[f"{source.upper()}"].value),
                )

                records = response['Items']
                if not records:
                    break

                with database.therapies.batch_writer(
                        overwrite_by_pkeys=['label_and_type', 'concept_id']) \
                        as batch:

                    for record in records:
                        batch.delete_item(
                            Key={
                                'label_and_type': record['label_and_type'],
                                'concept_id': record['concept_id']
                            }
                        )
        except ClientError as e:
            click.echo(e.response['Error']['Message'])


if __name__ == '__main__':
    CLI().update_normalizer_db()



"""This module defines the ChEMBL ETL methods."""
from .base import Base
from therapy import PROJECT_ROOT
from therapy.schemas import SourceName, NamespacePrefix, ApprovalStatus, \
    SourceMeta
import logging
import tarfile
import sqlite3
import os
import shutil

logger = logging.getLogger('therapy')
logger.setLevel(logging.DEBUG)


class ChEMBL(Base):
    """ETL the ChEMBL source into therapy.db."""

    def _extract_data(self, *args, **kwargs):
        """Extract data from the ChEMBL source."""
        logger.info('Extracting chembl_27.db...')
        if 'data_path' in kwargs:
            chembl_db = kwargs['data_path']
        else:
            chembl_db = self._src_data_dir / 'chembl_27.db'
        if not chembl_db.exists():
            chembl_archive = self._src_data_dir / 'chembl_27_sqlite.tar.gz'
            chembl_archive.parent.mkdir(exist_ok=True, parents=True)
            self._download_data()
            tar = tarfile.open(chembl_archive)
            tar.extractall(path=PROJECT_ROOT / 'data' / 'chembl')
            tar.close()

            # Remove unused directories and files
            chembl_27_dir = self._src_data_dir / 'chembl_27'
            temp_chembl = chembl_27_dir / 'chembl_27_sqlite' / 'chembl_27.db'
            chembl_db = self._src_data_dir / 'chembl_27.db'
            shutil.move(temp_chembl, chembl_db)
            os.remove(chembl_archive)
            shutil.rmtree(chembl_27_dir)
        conn = sqlite3.connect(chembl_db)
        conn.row_factory = sqlite3.Row
        self._conn = conn
        self._cursor = conn.cursor()
        assert chembl_db.exists()
        logger.info('Finished extracting chembl_27.db.')

    def _download_data(self, *args, **kwargs):
        """Download ChEMBL data from FTP."""
        logger.info(
            'Downloading ChEMBL v27, this will take a few minutes.')
        self._ftp_download('ftp.ebi.ac.uk',
                           'pub/databases/chembl/ChEMBLdb/releases/chembl_27',
                           self._src_data_dir,
                           'chembl_27_sqlite.tar.gz')

    def _transform_data(self, *args, **kwargs):
        """Transform SQLite data to JSON."""
        self._create_dictionary_synonyms_table()
        self._create_trade_names_table()
        self._create_temp_table()

        self._cursor.execute("DROP TABLE DictionarySynonyms;")
        self._cursor.execute("DROP TABLE TradeNames;")

        self._load_json()
        self._conn.commit()
        self._conn.close()

    def _create_dictionary_synonyms_table(self):
        """Create temporary table to store drugs and their synonyms."""
        create_dictionary_synonyms_table = f"""
            CREATE TEMPORARY TABLE DictionarySynonyms AS
            SELECT
                md.molregno,
                '{NamespacePrefix.CHEMBL.value}:'||md.chembl_id as chembl_id,
                md.pref_name,
                md.max_phase,
                md.withdrawn_flag,
                group_concat(
                    ms.synonyms, '||')as synonyms
            FROM molecule_dictionary md
            LEFT JOIN molecule_synonyms ms USING(molregno)
            GROUP BY md.molregno
            UNION ALL
            SELECT
                md.molregno,
                '{NamespacePrefix.CHEMBL.value}:'||md.chembl_id as chembl_id,
                md.pref_name,
                md.max_phase,
                md.withdrawn_flag,
                group_concat(
                    ms.synonyms, '||') as synonyms
            FROM molecule_synonyms ms
            LEFT JOIN molecule_dictionary md USING(molregno)
            WHERE md.molregno IS NULL
            GROUP BY md.molregno
        """
        self._cursor.execute(create_dictionary_synonyms_table)

    def _create_trade_names_table(self):
        """Create temporary table to store trade name data."""
        create_trade_names_table = """
            CREATE TEMPORARY TABLE TradeNames AS
            SELECT
                f.molregno,
                f.product_id,
                group_concat(
                    p.trade_name, '||') as trade_names
            FROM formulations f
            LEFT JOIN products p
                ON f.product_id = p.product_id
            GROUP BY f.molregno;
        """
        self._cursor.execute(create_trade_names_table)

    def _create_temp_table(self):
        """Create temporary table to store therapies data."""
        create_temp_table = """
            CREATE TEMPORARY TABLE temp(concept_id, label, approval_status,
                                        src_name, trade_names, aliases);
        """
        self._cursor.execute(create_temp_table)

        insert_temp = f"""
            INSERT INTO temp(concept_id, label, approval_status, src_name,
                             trade_names, aliases)
            SELECT
                ds.chembl_id,
                ds.pref_name,
                CASE
                    WHEN ds.withdrawn_flag
                        THEN '{ApprovalStatus.WITHDRAWN.value}'
                    WHEN ds.max_phase == 4
                        THEN '{ApprovalStatus.APPROVED.value}'
                    WHEN ds.max_phase == 0
                        THEN NULL
                    ELSE
                        '{ApprovalStatus.INVESTIGATIONAL.value}'
                END,
                '{SourceName.CHEMBL.value}',
                t.trade_names,
                ds.synonyms
            FROM DictionarySynonyms ds
            LEFT JOIN TradeNames t USING(molregno)
            GROUP BY ds.molregno
            UNION ALL
            SELECT
                ds.chembl_id,
                ds.pref_name,
                CASE
                    WHEN ds.withdrawn_flag
                        THEN '{ApprovalStatus.WITHDRAWN.value}'
                    WHEN ds.max_phase == 4
                        THEN '{ApprovalStatus.APPROVED.value}'
                    WHEN ds.max_phase == 0
                        THEN NULL
                    ELSE
                        '{ApprovalStatus.INVESTIGATIONAL.value}'
                END,
                '{SourceName.CHEMBL.value}',
                t.trade_names,
                ds.synonyms
            FROM TradeNames t
            LEFT JOIN DictionarySynonyms ds USING(molregno)
            WHERE ds.molregno IS NULL
            GROUP BY ds.molregno;
        """
        self._cursor.execute(insert_temp)

    def _load_json(self):
        """Load ChEMBL data into database."""
        chembl_data = """
            SELECT
                concept_id,
                label,
                approval_status,
                src_name,
                trade_names,
                aliases
            FROM temp;
        """
        result = [dict(row) for row in
                  self._cursor.execute(chembl_data).fetchall()]
        self._cursor.execute("DROP TABLE temp;")

        for record in result:
            for attr in ['aliases', 'trade_names']:
                if attr in record and record[attr]:
                    record[attr] = record[attr].split('||')
            self._load_therapy(record)

    def _load_meta(self, *args, **kwargs):
        """Add ChEMBL metadata."""
        metadata = SourceMeta(data_license='CC BY-SA 3.0',
                              data_license_url='https://creativecommons.org/licenses/by-sa/3.0/',  # noqa: E501
                              version='27',
                              data_url='http://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_27/',  # noqa: E501
                              rdp_url='http://reusabledata.org/chembl.html',
                              data_license_attributes={
                                  'non_commercial': False,
                                  'share_alike': True,
                                  'attribution': True
                              })
        params = dict(metadata)
        params['src_name'] = SourceName.CHEMBL.value
        self.database.metadata.put_item(Item=params)



"""A base class for extraction, transformation, and loading of data."""
from abc import ABC, abstractmethod
from ftplib import FTP
from pathlib import Path
from typing import List, Dict
from therapy import ACCEPTED_SOURCES, PROJECT_ROOT, ITEM_TYPES
from therapy.schemas import Drug
import logging


logger = logging.getLogger('therapy')
logger.setLevel(logging.DEBUG)


class Base(ABC):
    """The ETL base class."""

    def __init__(self, database, data_path=PROJECT_ROOT / 'data'):
        """Extract from sources.

        :param Database database: application database object
        :param Path data_path: path to app data directory
        """
        name = self.__class__.__name__.lower()
        self.database = database
        self._src_data_dir = data_path / name
        self.in_normalize = name in ACCEPTED_SOURCES
        self._added_ids = []

    def perform_etl(self) -> List[str]:
        """Public-facing method to begin ETL procedures on given data.

        Returned concept IDs can be passed to Merge method for computing
        merged concepts.

        :return: list of concept IDs which were successfully processed and
            uploaded.
        """
        self._extract_data()
        self._load_meta()
        self._transform_data()
        return self._added_ids

    def _download_data(self, *args, **kwargs):
        raise NotImplementedError

    def _ftp_download(self, host: str, data_dir: str, source_dir: Path,
                      data_fn: str) -> None:
        """Download data file from FTP site.
        :param str host: Source's FTP host name
        :param str data_dir: Data directory located on FTP site
        :param Path source_dir: Source's data directory
        :param str data_fn: Filename on FTP site to be downloaded
        """
        try:
            with FTP(host) as ftp:
                ftp.login()
                logger.debug(f"FTP login to {host} was successful")
                ftp.cwd(data_dir)
                with open(source_dir / data_fn, 'wb') as fp:
                    ftp.retrbinary(f'RETR {data_fn}', fp.write)
        except Exception as e:
            logger.error(f'FTP download failed: {e}')
            raise Exception(e)

    def _extract_data(self):
        """Get source file from data directory."""
        self._src_data_dir.mkdir(exist_ok=True, parents=True)
        src_file_prefix = f'{type(self).__name__.lower()}_'
        dir_files = [f for f in self._src_data_dir.iterdir()
                     if f.name.startswith(src_file_prefix)]
        if len(dir_files) == 0:
            self._download_data()
            dir_files = [f for f in self._src_data_dir.iterdir()
                         if f.name.startswith(src_file_prefix)]
        self._src_file = sorted(dir_files, reverse=True)[0]
        self._version = self._src_file.stem.split('_', 1)[1]

    @abstractmethod
    def _transform_data(self, *args, **kwargs):
        raise NotImplementedError

    def _load_therapy(self, therapy: Dict):
        """Load individual therapy record.

        :param Dict therapy: valid therapy object.
        """
        assert Drug(**therapy)
        concept_id = therapy['concept_id']

        for attr_type, item_type in ITEM_TYPES.items():
            if attr_type in therapy:
                value = therapy[attr_type]
                if value is not None and value != []:
                    if isinstance(value, str):
                        items = [value.lower()]
                    else:
                        therapy[attr_type] = list(set(value))
                        items = {item.lower() for item in value}
                        if attr_type in ['aliases', 'trade_names']:
                            # remove duplicates
                            if 'label' in therapy:
                                therapy[attr_type] = list(set(therapy[attr_type]) - {therapy['label']})  # noqa: E501

                            if attr_type == 'aliases' and \
                                    'trade_names' in therapy:
                                therapy[attr_type] = list(set(therapy[attr_type]) - set(therapy['trade_names']))  # noqa: E501

                            if len(items) > 20:
                                logger.debug(f"{concept_id} has > 20"
                                             f" {attr_type}.")
                                del therapy[attr_type]
                                continue

                    for item in items:
                        self.database.add_ref_record(item, concept_id,
                                                     item_type)
                else:
                    del therapy[attr_type]

        # handle detail fields
        approval_attrs = ('approval_status', 'approval_year', 'fda_indication')
        for field in approval_attrs:
            if approval_attrs in therapy and therapy[field] is None:
                del therapy[field]

        self.database.add_record(therapy)
        if self.in_normalize:
            self._added_ids.append(concept_id)

    @abstractmethod
    def _load_meta(self, *args, **kwargs):
        raise NotImplementedError



"""This module defines the Wikidata ETL methods."""
from .base import Base
from therapy import PROJECT_ROOT
from therapy.schemas import SourceName, NamespacePrefix, \
    SourceIDAfterNamespace, SourceMeta
import json
import logging
from pathlib import Path
from wikibaseintegrator import wbi_core
import datetime

logger = logging.getLogger('therapy')
logger.setLevel(logging.DEBUG)

# Prefixes for translating ID namespaces
IDENTIFIER_PREFIXES = {
    'casRegistry': NamespacePrefix.CASREGISTRY.value,
    'ChemIDplus': NamespacePrefix.CHEMIDPLUS.value,
    'pubchemCompound': NamespacePrefix.PUBCHEMCOMPOUND.value,
    'pubchemSubstance': NamespacePrefix.PUBCHEMSUBSTANCE.value,
    'chembl': NamespacePrefix.CHEMBL.value,
    'rxnorm': NamespacePrefix.RXNORM.value,
    'drugbank': NamespacePrefix.DRUGBANK.value,
    'wikidata': NamespacePrefix.WIKIDATA.value,
}


SPARQL_QUERY = """
    SELECT ?item ?itemLabel ?casRegistry ?pubchemCompound
           ?pubchemSubstance ?chembl
           ?rxnorm ?drugbank ?alias WHERE {
      ?item (wdt:P31/(wdt:P279*)) wd:Q12140.
      OPTIONAL {
        ?item skos:altLabel ?alias.
        FILTER((LANG(?alias)) = "en")
      }
      OPTIONAL { ?item p:P231 ?wds1.
                 ?wds1 ps:P231 ?casRegistry.
               }
      OPTIONAL { ?item p:P662 ?wds2.
                 ?wds2 ps:P662 ?pubchemCompound.
               }
      OPTIONAL { ?item p:P2153 ?wds3.
                 ?wds3 ps:P2153 ?pubchemSubstance.
               }
      OPTIONAL { ?item p:P592 ?wds4.
                 ?wds4 ps:P592 ?chembl
               }
      OPTIONAL { ?item p:P3345 ?wds5.
                 ?wds5 ps:P3345 ?rxnorm.
               }
      OPTIONAL { ?item p:P715 ?wds6.
                 ?wds6 ps:P715 ?drugbank
               }
      SERVICE wikibase:label {
        bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en".
      }
    }
"""


class Wikidata(Base):
    """Extract, transform, and load the Wikidata source into therapy.db."""

    def __init__(self,
                 database,
                 data_path: Path = PROJECT_ROOT / 'data'):
        """Initialize wikidata ETL class.

        :param therapy.database.Database: DB instance to use
        :param Path data_path: path to app data directory
        """
        super().__init__(database, data_path)

    def _extract_data(self):
        """Extract data from the Wikidata source."""
        self._src_data_dir.mkdir(exist_ok=True, parents=True)

        data = (wbi_core.FunctionsEngine.execute_sparql_query(
            SPARQL_QUERY))['results']['bindings']

        transformed_data = list()
        for item in data:
            params = dict()
            for attr in item:
                params[attr] = item[attr]['value']
            transformed_data.append(params)

        self._version = datetime.datetime.today().strftime('%Y%m%d')
        with open(f"{self._src_data_dir}/wikidata_{self._version}.json",
                  'w+') as f:
            json.dump(transformed_data, f)
        self._data_src = sorted(list(self._src_data_dir.iterdir()))[-1]
        logger.info('Successfully extracted Wikidata.')

    def _load_meta(self):
        """Add Wikidata metadata."""
        metadata = SourceMeta(src_name=SourceName.WIKIDATA.value,
                              data_license='CC0 1.0',
                              data_license_url='https://creativecommons.org/publicdomain/zero/1.0/',  # noqa: E501
                              version=self._version,
                              data_url=None,
                              rdp_url=None,
                              data_license_attributes={
                                  'non_commercial': False,
                                  'share_alike': False,
                                  'attribution': False
                              })
        params = dict(metadata)
        params['src_name'] = SourceName.WIKIDATA.value
        self.database.metadata.put_item(Item=params)

    def _transform_data(self):
        """Transform the Wikidata source data."""
        from therapy import XREF_SOURCES
        with open(self._data_src, 'r') as f:
            records = json.load(f)

            items = dict()

            for record in records:
                record_id = record['item'].split('/')[-1]
                concept_id = f"{NamespacePrefix.WIKIDATA.value}:{record_id}"
                if concept_id not in items.keys():
                    item = dict()
                    item['label_and_type'] = f"{concept_id.lower()}##identity"
                    item['item_type'] = 'identity'
                    item['concept_id'] = concept_id
                    item['src_name'] = SourceName.WIKIDATA.value

                    xrefs = list()
                    associated_with = list()
                    for key in IDENTIFIER_PREFIXES.keys():
                        if key in record.keys():
                            ref = record[key]

                            if key.upper() == 'CASREGISTRY':
                                key = SourceName.CHEMIDPLUS.value

                            if key.upper() in XREF_SOURCES:
                                if key != 'chembl':
                                    fmted_xref = \
                                        f"{IDENTIFIER_PREFIXES[key]}:{SourceIDAfterNamespace[key.upper()].value}{ref}"  # noqa: E501
                                else:
                                    fmted_xref = \
                                        f"{IDENTIFIER_PREFIXES[key]}:{ref}"
                                xrefs.append(fmted_xref)
                            else:
                                fmted_assoc = f"{IDENTIFIER_PREFIXES[key]}:" \
                                              f"{ref}"
                                associated_with.append(fmted_assoc)
                    item['xrefs'] = xrefs
                    item['associated_with'] = associated_with
                    if 'itemLabel' in record.keys():
                        item['label'] = record['itemLabel']
                    items[concept_id] = item
                if 'alias' in record.keys():
                    if 'aliases' in items[concept_id].keys():
                        items[concept_id]['aliases'].append(record['alias'])
                    else:
                        items[concept_id]['aliases'] = [record['alias']]

        for item in items.values():
            self._load_therapy(item)



"""ETL methods for NCIt source"""
from .base import Base
from therapy import PROJECT_ROOT
from therapy.schemas import SourceName, NamespacePrefix, Therapy, SourceMeta
import logging
import owlready2 as owl
from owlready2.entity import ThingClass
from typing import Set
from pathlib import Path
import requests
import zipfile
from os import remove, rename

logger = logging.getLogger('therapy')
logger.setLevel(logging.DEBUG)


class NCIt(Base):
    """Core NCIt ETL class.

    Extracting both:
     * NCIt classes with semantic_type "Pharmacologic Substance" but not
       Retired_Concept
     * NCIt classes that are subclasses of C1909 (Pharmacologic Substance)
    """

    def __init__(self,
                 database,
                 src_dir: str = "https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/archive/2020/20.09d_Release/",  # noqa F401
                 src_fname: str = "Thesaurus_20.09d.OWL.zip",
                 data_path: Path = PROJECT_ROOT / 'data',
                 chemidplus_path: Path = PROJECT_ROOT / 'data' / 'chemidplus'):
        """Override base class init method. Call ETL methods.

        :param therapy.database.Database database: app database instance
        :param str src_dir: URL of remote directory containing source input
        :param str src_fname: filename for source file within source directory
        :param pathlib.Path data_path: path to local NCIt data directory
        """
        super().__init__(database, data_path)
        self._SRC_DIR = src_dir
        self._SRC_FNAME = src_fname
        self._chemidplus_path = chemidplus_path

    def _download_data(self):
        """Download NCI thesaurus source file for loading into normalizer."""
        logger.info('Downloading NCI Thesaurus...')
        url = self._SRC_DIR + self._SRC_FNAME
        zip_path = self._src_data_dir / 'ncit.zip'
        response = requests.get(url, stream=True)
        handle = open(zip_path, "wb")
        for chunk in response.iter_content(chunk_size=512):
            if chunk:
                handle.write(chunk)
        handle.close()
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            zip_ref.extractall(self._src_data_dir)
        remove(zip_path)
        version = self._SRC_DIR.split('/')[-2].split('_')[0]
        rename(self._src_data_dir / 'Thesaurus.owl', self._src_data_dir / f'ncit_{version}.owl')  # noqa: E501
        logger.info('Finished downloading NCI Thesaurus')

    def _get_desc_nodes(self, node: ThingClass,
                        uq_nodes: Set[ThingClass]) -> Set[ThingClass]:
        """Create set of unique subclasses of node parameter.
        Should be originally called on ncit:C1909: Pharmacologic Substance.

        :param owlready2.entity.ThingClass node: concept node to either
            retrieve descendants of, or to normalize and add to DB
        :param Set[owlready2.entity.ThingClass] uq_nodes: set of unique class
            nodes found so far from recursive tree exploration
        :return: the uq_nodes set, updated with any class nodes found from
            recursive exploration of this branch of the class tree
        :rtype: Set[owlready2.entity.ThingClass]
        """
        children = node.descendants()
        if children:
            for child_node in children:
                if child_node is not node:
                    uq_nodes.add(child_node)
                    self._get_desc_nodes(child_node, uq_nodes)
        return uq_nodes

    def _get_typed_nodes(self, uq_nodes: Set[ThingClass],
                         ncit: owl.namespace.Ontology) -> Set[ThingClass]:
        """Get all nodes with semantic_type Pharmacologic Substance

        :param Set[owlready2.entity.ThingClass] uq_nodes: set of unique class
            nodes found so far.
        :param owl.namespace.Ontology ncit: owlready2 Ontology instance for
            NCI Thesaurus.
        :return: uq_nodes, with the addition of all classes found to have
            semantic_type Pharmacologic Substance and not of type
            Retired_Concept
        :rtype: Set[owlready2.entity.ThingClass]
        """
        graph = owl.default_world.as_rdflib_graph()

        query_str = '''SELECT ?x WHERE {
            ?x <http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#P106>
            "Pharmacologic Substance"
        }'''
        typed_results = set(graph.query(query_str))

        retired_query_str = '''SELECT ?x WHERE {
            ?x <http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#P310>
            "Retired_Concept"
        }
        '''
        retired_results = set(graph.query(retired_query_str))

        typed_results = typed_results - retired_results

        for result in typed_results:
            # parse result as URI and get ThingClass object back from NCIt
            class_object = ncit[result[0].toPython().split('#')[1]]
            uq_nodes.add(class_object)
        return uq_nodes

    def _transform_data(self):
        """Get data from file and construct objects for loading"""
        ncit = owl.get_ontology(self._src_file.absolute().as_uri())
        ncit.load()
        uq_nodes = {ncit.C49236}  # add Therapeutic Procedure
        uq_nodes = self._get_desc_nodes(ncit.C1909, uq_nodes)
        uq_nodes = self._get_typed_nodes(uq_nodes, ncit)
        for node in uq_nodes:
            concept_id = f"{NamespacePrefix.NCIT.value}:{node.name}"
            if node.P108:
                label = node.P108.first()
            else:
                label = None
            aliases = node.P90
            if label and aliases and label in aliases:
                aliases.remove(label)

            xrefs = []
            associated_with = []
            if node.P207:
                associated_with.append(f"{NamespacePrefix.UMLS.value}:"
                                       f"{node.P207.first()}")
            if node.P210:
                xrefs.append(f"{NamespacePrefix.CASREGISTRY.value}:"
                             f"{node.P210.first()}")
            if node.P319:
                associated_with.append(f"{NamespacePrefix.FDA.value}:"
                                       f"{node.P319.first()}")
            if node.P320:
                associated_with.append(f"{NamespacePrefix.ISO.value}:"
                                       f"{node.P320.first()}")
            if node.P368:
                associated_with.append(f"{NamespacePrefix.CHEBI.value}:"
                                       f"{node.P368.first()}")
            params = {
                'concept_id': concept_id,
                'label': label,
                'aliases': aliases,
                'xrefs': xrefs,
                'associated_with': associated_with
            }
            assert Therapy(**params)
            self._load_therapy(params)

    def _load_meta(self):
        """Load metadata"""
        metadata = SourceMeta(data_license="CC BY 4.0",
                              data_license_url="https://creativecommons.org/licenses/by/4.0/legalcode",  # noqa F401
                              version=self._version,
                              data_url=self._SRC_DIR,
                              rdp_url='http://reusabledata.org/ncit.html',
                              data_license_attributes={
                                  'non_commercial': False,
                                  'share_alike': False,
                                  'attribution': True
                              })
        params = dict(metadata)
        params['src_name'] = SourceName.NCIT.value
        self.database.metadata.put_item(Item=params)



"""Main application for FastAPI"""
from therapy import __version__
from therapy.query import QueryHandler, InvalidParameterException
from therapy.schemas import SearchService, NormalizationService
from fastapi import FastAPI, HTTPException, Query
from fastapi.openapi.utils import get_openapi
import html
from typing import Optional

query_handler = QueryHandler()
app = FastAPI(docs_url='/therapy', openapi_url='/therapy/openapi.json')


def custom_openapi():
    """Generate custom fields for OpenAPI response"""
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="The VICC Therapy Normalizer",
        version=__version__,
        openapi_version="3.0.3",
        description="Normalize drugs and other therapy terms.",
        routes=app.routes
    )
#    openapi_schema['info']['license'] = {  # TODO
#        "name": "Name-of-license",
#        "url": "http://www.to-be-determined.com"
#    }
    openapi_schema['info']['contact'] = {  # TODO
        "name": "Alex H. Wagner",
        "email": "Alex.Wagner@nationwidechildrens.org"
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


app.openapi = custom_openapi

# endpoint description text
get_matches_summary = ("Given query, provide highest matches from "
                       "each source.")
response_descr = "A response to a validly-formed query."
q_descr = "Therapy to search."
keyed_descr = ("If true, return response as key-value pairs of "
               "sources to source matches.")
incl_descr = ("Comma-separated list of source names to include in "
              "response. Will exclude all other sources. Will return HTTP "
              "status code 422: Unprocessable Entity if both 'incl' and "
              "'excl' parameters are given.")
excl_descr = ("Comma-separated list of source names to exclude in "
              "response. Will include all other sources. Will return HTTP "
              "status code 422: Unprocessable Entity if both 'incl' and"
              "'excl' parameters are given.")
search_description = ("For each source, return strongest-match concepts "
                      "for query string provided by user")
normalize_description = ("Return merged strongest-match concept for query "
                         "string provided by user.")


@app.get("/therapy/search",
         summary=get_matches_summary,
         operation_id="getQueryResponse",
         response_description=response_descr,
         response_model=SearchService,
         description=search_description)
def search(q: str = Query(..., description=q_descr),
           keyed: Optional[bool] = Query(False, description=keyed_descr),
           incl: Optional[str] = Query(None, description=incl_descr),
           excl: Optional[str] = Query(None, description=excl_descr)):
    """For each source, return strongest-match concepts for query string
    provided by user.

        :param q: therapy search term
        :param keyed: if true, response is structured as key/value pair of
            sources to source match lists.
        :param incl: comma-separated list of sources to include, with all
            others excluded. Raises HTTPException if both `incl` and
            `excl` are given.
        :param excl: comma-separated list of sources exclude, with all others
            included. Raises HTTPException if both `incl` and `excl` are given.
        :returns: JSON response with matched records and source metadata
    """
    try:
        response = query_handler.search_sources(html.unescape(q), keyed=keyed,
                                                incl=incl, excl=excl)
    except InvalidParameterException as e:
        raise HTTPException(status_code=422, detail=str(e))
    return response


merged_matches_summary = "Given query, provide merged normalized record."
merged_response_descr = "A response to a validly-formed query."
merged_q_descr = "Therapy to normalize."


@app.get("/therapy/normalize",
         summary=merged_matches_summary,
         operation_id="getQuerymergedResponse",
         response_description=merged_response_descr,
         response_model=NormalizationService,
         description=normalize_description)
def normalize(q: str = Query(..., description=merged_q_descr)):
    """Return merged strongest-match concept for query string provided by
    user.

    :param q: therapy search term
    """
    try:
        response = query_handler.search_groups(html.unescape(q))
    except InvalidParameterException as e:
        raise HTTPException(status_code=422, detail=str(e))
    return response



"""ETL methods for ChemIDPlus® source.

Courtesy of the U.S. National Library of Medicine.
"""
from .base import Base
from therapy import PROJECT_ROOT
from therapy.schemas import Drug, NamespacePrefix, SourceMeta, SourceName, \
    DataLicenseAttributes
from pathlib import Path
import xml.etree.ElementTree as ET
import logging
import re


logger = logging.getLogger('therapy')
logger.setLevel(logging.DEBUG)


TAGS_REGEX = r' \[.*\]'


class ChemIDplus(Base):
    """Core ChemIDplus ETL class."""

    def __init__(self,
                 database,
                 data_path: Path = PROJECT_ROOT / 'data',
                 src_server: str = 'ftp.nlm.nih.gov',
                 src_dir_path: str = 'nlmdata/.chemidlease/',
                 src_fname: str = 'CurrentChemID.xml'):
        """Initialize class instance.

        :param Database database: application database object
        :param Path data_path: path to app data directory
        :param str src_server: The NLM domain
        :param str src_dir_path: The directory to the chemidplus release
        :param str src_fname: name of file as stored in src_dir.

        If the source file is provided locally in the data_path directory,
        it's unnecessary to provide `src_dir` and `src_fname` args.
        """
        super().__init__(database, data_path)
        self._src_server = src_server
        self._src_dir_path = src_dir_path
        self._src_fname = src_fname

    def _download_data(self):
        """Download source data from default location."""
        logger.info('Downloading ChemIDplus data...')
        outfile_path = self._src_data_dir / self._src_fname

        self._ftp_download(self._src_server,
                           self._src_dir_path,
                           self._src_data_dir,
                           self._src_fname)

        parser = ET.iterparse(outfile_path, ('start', 'end'))
        date = next(parser)[1].attrib['date']
        version = date.replace('-', '')
        outfile_path.rename(self._src_data_dir / f'chemidplus_{version}.xml')
        logger.info('Finished downloading ChemIDplus data')

    def _extract_data(self):
        """Acquire ChemIDplus dataset.

        :arg pathlib.Path data_path: directory containing source data
        """
        self._src_data_dir.mkdir(exist_ok=True, parents=True)
        dir_files = list(self._src_data_dir.iterdir())

        if len(dir_files) == 0:
            file = self._get_file()
        else:
            file = sorted([f for f in dir_files
                           if f.name.startswith('chemidplus')])
            if not file:
                file = self._get_file()

        self._data_src = file[-1]
        self._version = self._data_src.stem.split('_')[1]

    def _get_file(self):
        self._download_data()
        dir_files = list(self._src_data_dir.iterdir())
        return sorted([f for f in dir_files
                       if f.name.startswith('chemidplus')])

    @staticmethod
    def parse_xml(path: Path, tag: str):
        """Parse XML file and retrieve elements with matching tag value.
        :param Path path: path to XML file
        :param str tag: XML tag
        :return: generator yielding elements of corresponding tag
        """
        context = iter(ET.iterparse(path, events=('start', 'end')))
        _, root = next(context)
        for event, elem in context:
            if event == 'end' and elem.tag == tag:
                yield elem
                root.clear()

    def _transform_data(self):
        """Open dataset and prepare for loading into database."""
        parser = self.parse_xml(self._data_src, 'Chemical')
        for chemical in parser:
            if 'displayName' not in chemical.attrib:
                continue

            # initial setup and get label
            display_name = chemical.attrib['displayName']
            if not display_name or not re.search(TAGS_REGEX, display_name):
                continue
            label = re.sub(TAGS_REGEX, '', display_name)
            params = {
                'label': label
            }

            # get concept ID
            reg_no = chemical.find('NumberList').find("CASRegistryNumber")
            if not reg_no:
                continue
            params['concept_id'] = \
                f'{NamespacePrefix.CASREGISTRY.value}:{reg_no.text}'

            # get aliases
            aliases = []
            label_l = label.lower()
            name_list = chemical.find('NameList')
            if name_list:
                for name in name_list.findall('NameOfSubstance'):
                    text = name.text
                    if text != display_name and text.lower() != label_l:
                        aliases.append(re.sub(TAGS_REGEX, '', text))
            params['aliases'] = aliases

            # get xrefs and associated_with
            params['xrefs'] = []
            params['associated_with'] = []
            locator_list = chemical.find('LocatorList')
            if locator_list:
                for loc in locator_list.findall('InternetLocator'):
                    if loc.text == 'DrugBank':
                        db = f'{NamespacePrefix.DRUGBANK.value}:' \
                             f'{loc.attrib["url"].split("/")[-1]}'
                        params['xrefs'].append(db)
                    elif loc.text == 'FDA SRS':
                        fda = f'{NamespacePrefix.FDA.value}:' \
                              f'{loc.attrib["url"].split("/")[-1]}'
                        params['associated_with'].append(fda)

            # double-check and load full record
            assert Drug(**params)
            self._load_therapy(params)

    def _load_meta(self):
        """Add source metadata."""
        meta = SourceMeta(data_license="custom",
                          data_license_url="https://www.nlm.nih.gov/databases/download/terms_and_conditions.html",  # noqa: E501
                          version=self._version,
                          data_url="ftp://ftp.nlm.nih.gov/nlmdata/.chemidlease/",  # noqa: E501
                          rdp_url=None,
                          data_license_attributes=DataLicenseAttributes(
                              non_commercial=False,
                              share_alike=False,
                              attribution=True
                          ))
        item = dict(meta)
        item['src_name'] = SourceName.CHEMIDPLUS.value
        self.database.metadata.put_item(Item=item)



"""This module defines the DrugBank ETL methods."""
from therapy import DownloadException
from therapy.schemas import SourceName, SourceMeta, NamespacePrefix
from therapy.etl.base import Base
import logging
import csv
import bs4
import requests
import re
import zipfile
import shutil
from io import BytesIO

logger = logging.getLogger('therapy')
logger.setLevel(logging.DEBUG)


class DrugBank(Base):
    """ETL the DrugBank source into therapy.db."""

    def _download_data(self):
        """Download DrugBank source data."""
        logger.info("Downloading DrugBank source data...")

        # get newest version number
        r = requests.get('https://go.drugbank.com/release_notes')
        if r.status_code == 200:
            soup = bs4.BeautifulSoup(r.content, features='lxml')
        else:
            logger.error(f'DrugBank version fetch failed with status code: '
                         f'{r.status_code}')
            raise DownloadException
        most_recent = soup.find('div', {'class': 'card-header'})
        version = re.search(r'[0-9]+\.[0-9]+\.[0-9]+',
                            most_recent.contents[0]).group()
        url = f'https://go.drugbank.com/releases/{version.replace(".", "-")}/downloads/all-drugbank-vocabulary'  # noqa: E501

        # download file
        r = requests.get(url)
        if r.status_code == 200:
            zip_file = zipfile.ZipFile(BytesIO(r.content))
        else:
            logger.error("DrugBank download failed with status code:"
                         f" {r.status_code}")
            raise requests.HTTPError(r.status_code)

        # unpack file
        temp_dir = self._src_data_dir / 'temp_drugbank'
        zip_file.extractall(temp_dir)
        temp_file = temp_dir / 'drugbank vocabulary.csv'
        csv_file = self._src_data_dir / f'drugbank_{version}.csv'
        shutil.move(temp_file, csv_file)
        shutil.rmtree(temp_dir)
        logger.info("DrugBank source data download complete.")

    def _load_meta(self):
        """Add DrugBank metadata."""
        meta = {
            'data_license': 'CC0 1.0',
            'data_license_url': 'https://creativecommons.org/publicdomain/zero/1.0/',  # noqa: E501
            'version': self._version,
            'data_url': 'https://go.drugbank.com/releases/latest#open-data',
            'rdp_url': 'http://reusabledata.org/drugbank.html',
            'data_license_attributes': {
                'non_commercial': False,
                'share_alike': False,
                'attribution': False,
            },
        }
        assert SourceMeta(**meta)
        meta['src_name'] = SourceName.DRUGBANK.value
        self.database.metadata.put_item(Item=meta)

    def _transform_data(self):
        """Transform the DrugBank source."""
        with open(self._src_file, 'r') as file:
            reader = csv.reader(file)
            next(reader)  # skip header
            for row in reader:
                # get concept ID
                params = {
                    'concept_id': f'{NamespacePrefix.DRUGBANK.value}:{row[0]}',
                }

                # get label
                label = row[2]
                if label:
                    params['label'] = label

                # get aliases
                aliases = [
                    a for a in row[1].split(' | ') + row[5].split(' | ') if a
                ]
                if aliases:
                    params['aliases'] = aliases

                # get CAS reference
                cas_ref = row[3]
                if cas_ref:
                    params['xrefs'] = [
                        f'{NamespacePrefix.CHEMIDPLUS.value}:{row[3]}'
                    ]

                params['associated_with'] = []
                # get inchi key
                if len(row) >= 7:
                    inchi_key = row[6]
                    if inchi_key:
                        inchi_id = f'{NamespacePrefix.INCHIKEY.value}:{inchi_key}'  # noqa: E501
                        params['associated_with'].append(inchi_id)
                # get UNII id
                unii = row[4]
                if unii:
                    unii_id = f'{NamespacePrefix.UNII.value}:{unii}'
                    params['associated_with'].append(unii_id)

                self._load_therapy(params)
